import java.util.HashMap;
import java.util.List;

/**
 * <p>
 * 	A tree-based mutex.
 * </p>
 * <p>
 * 	The mutual exclusion algorithm uses a binary tree where each internal (non-leaf) node represents a critical section shared by its descendants. The threads are at the leaves of the tree. The root of the tree is the main critical section shared by all the threads.
 * <p>
 * 	To enter the main critical section, a thread starts at its leaf in the tree. The thread is required to traverse the path from its leaf up to the root, entering all the critical sections on its path. Upon exiting the critical section, the thread traverses this path in reverse, this time leaving all the critical sections on its path.
 * </p>
 * <p>
 * 	At each internal node, there is a maximum of two threads competing against each other to enter the node's critical section. The Peterson's algorithm for 2 threads is be used to implement the critical section of an internal node.
 * </p>
 */
public class TreeBasedMutex {
	/**
	 * Create a tree-based mutex.
	 * 
	 * @param threads	The list of threads that use this mutex.
	 */
	public TreeBasedMutex(List<Thread> threads) {
		/* Initialize the linear identifiers for threads. */
		threadLinearIdentifiers = new HashMap<Long, Integer>();
		for (int i = 0; i < threads.size(); i++) {
			threadLinearIdentifiers.put(threads.get(i).getId(), i);
		}
		
		/* Initialize the Peterson flags. */
		lowerSubIntervalFlags = new boolean[threadLinearIdentifiers.size()];
		upperSubIntervalFlags = new boolean[threadLinearIdentifiers.size()];
		victim = new int[threadLinearIdentifiers.size()];
	}
	
	/**
	 * Wait until this mutex is locked.
	 */
	public void lock() {
		enterInterval (0, threadLinearIdentifiers.size() - 1);
		System.out.print("Thread " + Thread.currentThread().getId() + " obtained the lock.\n");
	}
	
	/**
	 * Unlock this mutex.
	 */
	public void unlock() {
		leaveInterval (0, threadLinearIdentifiers.size() - 1);
		System.out.print("Thread " + Thread.currentThread().getId() + " released the lock.\n");
	}

	/**
	 * Enter the critical section of the thread interval between <code>lowerBound</code> and <code>upperBound</code>.
	 * 
	 * @param lowerBound	The lower end of the interval.
	 * @param upperBound	The upper end of the interval.
	 */
	public void enterInterval (int lowerBound, int upperBound) {
		int median;
		boolean isLowerSubInterval;
		
		if (lowerBound != upperBound) {
			/* The interval contains more than one thread. */
			
			/* Evaluate the median, the lower and the upper sub interval. */
			median = lowerBound + (upperBound - lowerBound) / 2;
			isLowerSubInterval = threadLinearIdentifiers.get(Thread.currentThread().getId()) <= median;
			
			/* Recursively perform the operation on the subinterval. */
			if (isLowerSubInterval) {
				enterInterval(lowerBound, median);
			} else {
				enterInterval(median + 1, upperBound);
			}
			
			/* Start Peterson's algorithm to determine who should enter the critical section of the current interval. */
			if (isLowerSubInterval) {
				lowerSubIntervalFlags[median] = true;
			} else {
				upperSubIntervalFlags[median] = true;
			}
			victim[median] = threadLinearIdentifiers.get(Thread.currentThread().getId());
			if (isLowerSubInterval) {
				while (upperSubIntervalFlags[median] == true && victim[median] == threadLinearIdentifiers.get(Thread.currentThread().getId())) {
					/* Instead of a true busy loop, we allow the current thread to pause while waiting. This improves the performance in a busy loop. */
					Thread.yield();
				}
			} else {
				while (lowerSubIntervalFlags[median] == true && victim[median] == threadLinearIdentifiers.get(Thread.currentThread().getId())) {
					/* Instead of a true busy loop, we allow the current thread to pause while waiting. This improves the performance in a busy loop. */
					Thread.yield();
				}
			}
		}
	}

	/**
	 * Leave the critical section of the thread interval between <code>lowerBound</code> and <code>upperBound</code>.
	 * 
	 * @param lowerBound	The lower end of the interval.
	 * @param upperBound	The upper end of the interval.
	 */
	public void leaveInterval (int lowerBound, int upperBound) {
		int median;
		boolean isLowerSubInterval;
		
		if (lowerBound != upperBound) {
			/* The interval contains more than one thread. */
			
			
			/* Evaluate the median, the lower and the upper sub interval. */
			median = lowerBound + (upperBound - lowerBound) / 2;
			isLowerSubInterval = threadLinearIdentifiers.get(Thread.currentThread().getId()) <= median;
			
			/* Start Peterson's algorithm to leave the critical section of the current interval. */
			if (isLowerSubInterval) {
				lowerSubIntervalFlags[median] = false;
			} else {
				upperSubIntervalFlags[median] = false;
			}
			
			/* Recursively perform the operation on the subinterval. */
			if (isLowerSubInterval) {
				leaveInterval(lowerBound, median);
			} else {
				leaveInterval(median + 1, upperBound);
			}
		}
	}
	
	/**
	 * The Peterson algorithm flag for a particular lower subinterval of an interval i. The array is indexed by the median of the interval i. 
	 */
	private boolean[] lowerSubIntervalFlags;
	/**
	 * The Peterson algorithm flag for a particular higher subinterval of an interval i. The array is indexed by the median of the interval i.
	 */
	private boolean[] upperSubIntervalFlags;
	/**
	 * The Peterson algorithm victim variable for a particular interval i. The array is indexed by the median of the interval i.
	 */
	private int[] victim;
	/**
	 * The mapping of threads to a sequence from 0 to the number of threads - 1.
	 * This mapping is necessary to use thread identifications for array indices.
	 */
	private HashMap<Long, Integer> threadLinearIdentifiers;
}
